/* begincopyright
  Copyright (c) 1988 Xerox Corporation. All rights reserved.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted. Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws. This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification. Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    PCR Coordinator
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA
  endcopyright */
/*
 *  Copyright Xerox, Inc., 1988.
 *     Permission to recopy for non-commercial purposes is granted, 
 *     provided this copyright remains intact.
 */

#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <stdio.h>
#include <a.out.h>

#define STR(x) ((char *)((x)->n_un.n_strx + (long)stringtab))

int quiet = 0;
struct exec header;
int num_syms;
struct nlist *symtab, *find_sym();
char *stringtab;

main(argc, argv)
char **argv;
{
  int i;
  char *filename;
  FILE *file;
  int str_len;
  int file_length;
  struct stat the_stat_buffer;
  struct nlist *sym;

  /* check args */
  if (argc < 3) {
    printf("%s [-q] filename list-of-syms-to-unhide.\n\twhere -q means don't complain about missing or already unhidden symbols.\n", argv[0]);
    exit(1);
  }

  /* check for arguments */
  if (strcmp(argv[1], "-q") == 0) {
    quiet = 1;
    argv += 1;
    argc -= 1;
  };
  
  /* try to get file */
  filename = argv[1];
  if ((file = fopen(filename, "r+")) == NULL) {
    printf("ldhide: Cannot open file '%s'.\n", filename);
    exit(1);
  }

  /* get basic information about file */
  fstat(fileno(file), &the_stat_buffer);
  file_length = the_stat_buffer.st_size;
  fread(&header, sizeof(struct exec), 1, file);

  /* get symbol table */
  num_syms = header.a_syms / sizeof(struct nlist);
  if (num_syms <= 0) {
    printf("ldhide: No symbol table.\n");
    exit(1);
  }
  symtab = (struct nlist *)calloc(num_syms, sizeof(struct nlist));
  fseek(file, N_SYMOFF(header), 0);
  fread(symtab, sizeof(struct nlist), num_syms, file);
  
  /* get string table for symbol table */
  str_len = file_length - N_STROFF(header);
  if (str_len <= 0) {
    printf("ldhide: No string table.\n");
    exit(1);
  }
  stringtab = (char *)calloc(str_len, 1);
  fseek(file, N_STROFF(header), 0);
  fread(stringtab, 1, str_len, file);

  /* look for symbols to hide, turn on the EXT bit */
  for (i = 2; i < argc; i++) {
    if ((sym = find_sym(argv[i]))) {
      if ((sym->n_type & N_EXT)) {
	if (!quiet) {
	  printf("ldhide: Symbol '%s' already visible.\n", argv[i]);
	}
      } else if ((sym->n_type & N_TYPE) == N_UNDF) {
	printf("ldhide: Symbol '%s' is undefined: probably not a good idea to unhide it.\n", argv[i]);
	} else {
	sym->n_type |= N_EXT;
      }
    } else {
      if (!quiet) {
	printf("ldhide: Symbol '%s' not found.\n", argv[i]);
      }
    }
  }

  /* write out new symtab */
  fseek(file, N_SYMOFF(header), 0);
  fwrite(symtab, sizeof(struct nlist), num_syms, file);
  exit(0);
}

/*
 * proc to look for a symbol
 */
struct nlist *
find_sym(s)
char *s;
{
  int i;
  for (i=0; i < num_syms; i++) {
    if (strcmp(STR(&symtab[i]), s) == 0)
      return &symtab[i];
  }
  return 0;
}
